#!/usr/bin/env python3
"""
Subdomain Scanner
DNS enumeration and subdomain discovery tool.
Parameters:
    - domain: Target domain name
    - wordlist: Path to subdomain wordlist
    - resolvers: List of DNS resolvers to use
Requires: dnspython package
"""
import dns.resolver
import dns.query
import dns.zone
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from concurrent.futures import ThreadPoolExecutor
import socket
import ipaddress

class SubdomainScanner:
    def __init__(self, domain, wordlist=None, resolvers=None, threads=50):
        self.domain = domain
        self.wordlist = wordlist
        self.resolvers = resolvers or ['8.8.8.8', '8.8.4.4']
        self.threads = threads
        self.results_queue = queue.Queue()
        self.setup_logging()
        self.setup_resolver()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def setup_resolver(self):
        self.resolver = dns.resolver.Resolver()
        self.resolver.nameservers = [socket.gethostbyname(r) if not 
                                   self.is_ip(r) else r 
                                   for r in self.resolvers]
        self.resolver.timeout = 2
        self.resolver.lifetime = 2

    def is_ip(self, address):
        try:
            ipaddress.ip_address(address)
            return True
        except ValueError:
            return False

    def load_wordlist(self):
        if not self.wordlist:
            # Default small wordlist for common subdomains
            return ['www', 'mail', 'ftp', 'localhost', 'webmail', 'smtp', 'pop',
                   'ns1', 'ns2', 'ns3', 'ns4', 'webdisk', 'pop3', 'cpanel',
                   'whm', 'autodiscover', 'autoconfig', 'news', 'cp', 'blog',
                   'dev', 'admin', 'test', 'portal', 'dns1', 'dns2', 'dns3',
                   'dns4', 'direct', 'direct1', 'direct2', 'vpn', 'remote']

        try:
            with open(self.wordlist, 'r') as f:
                return [line.strip() for line in f if line.strip()]
        except Exception as e:
            self.logger.error(f"Error loading wordlist: {str(e)}")
            return []

    def query_domain(self, subdomain):
        full_domain = f"{subdomain}.{self.domain}"
        try:
            answers = self.resolver.resolve(full_domain, 'A')
            ips = [str(answer) for answer in answers]
            
            # Try to get additional records
            mx_records = []
            txt_records = []
            ns_records = []
            
            try:
                mx = self.resolver.resolve(full_domain, 'MX')
                mx_records = [str(x.exchange) for x in mx]
            except:
                pass
                
            try:
                txt = self.resolver.resolve(full_domain, 'TXT')
                txt_records = [str(x) for x in txt]
            except:
                pass
                
            try:
                ns = self.resolver.resolve(full_domain, 'NS')
                ns_records = [str(x) for x in ns]
            except:
                pass

            result = {
                'subdomain': subdomain,
                'domain': full_domain,
                'ips': ips,
                'mx_records': mx_records,
                'txt_records': txt_records,
                'ns_records': ns_records
            }
            
            self.results_queue.put(result)
            return True
            
        except Exception as e:
            if not isinstance(e, (dns.resolver.NXDOMAIN, dns.resolver.NoAnswer)):
                self.logger.debug(f"Error querying {full_domain}: {str(e)}")
            return False

    def check_zone_transfer(self):
        try:
            ns_answer = self.resolver.resolve(self.domain, 'NS')
            for ns in ns_answer:
                try:
                    zone = dns.zone.from_xfr(dns.query.xfr(str(ns), self.domain))
                    names = zone.nodes.keys()
                    return [str(name) + '.' + self.domain for name in names]
                except:
                    continue
        except:
            pass
        return []

    def scan_subdomains(self):
        self.logger.info(f"Starting subdomain scan of {self.domain}")
        
        # First try zone transfer
        zone_transfer_results = self.check_zone_transfer()
        if zone_transfer_results:
            self.logger.info("Successfully performed zone transfer!")
            for domain in zone_transfer_results:
                self.query_domain(domain.rstrip(f".{self.domain}"))
        
        # Then try bruteforce
        wordlist = self.load_wordlist()
        self.logger.info(f"Loaded {len(wordlist)} subdomain names")
        
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            executor.map(self.query_domain, wordlist)

    def save_results(self, output_file):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        scan_results = {
            'timestamp': datetime.now().isoformat(),
            'domain': self.domain,
            'resolvers': self.resolvers,
            'subdomains_found': len(results),
            'results': sorted(results, key=lambda x: x['domain'])
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(scan_results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_summary(self):
        results = []
        while not self.results_queue.empty():
            results.append(self.results_queue.get())

        self.logger.info("\nScan Summary:")
        self.logger.info(f"Total subdomains found: {len(results)}")

        # Count unique IPs
        all_ips = set()
        for result in results:
            all_ips.update(result['ips'])
        self.logger.info(f"Unique IP addresses: {len(all_ips)}")

        # Group by IP
        domains_by_ip = {}
        for result in results:
            for ip in result['ips']:
                if ip not in domains_by_ip:
                    domains_by_ip[ip] = []
                domains_by_ip[ip].append(result['domain'])

        # Show interesting findings
        self.logger.info("\nInteresting findings:")
        
        # Domains sharing IPs
        shared_ips = {ip: domains for ip, domains in domains_by_ip.items() 
                     if len(domains) > 1}
        if shared_ips:
            self.logger.info("\nDomains sharing IP addresses:")
            for ip, domains in shared_ips.items():
                self.logger.info(f"\n{ip}:")
                for domain in domains:
                    self.logger.info(f"  - {domain}")

        # Domains with mail servers
        mail_servers = [r for r in results if r['mx_records']]
        if mail_servers:
            self.logger.info("\nDomains with mail servers:")
            for result in mail_servers:
                self.logger.info(f"{result['domain']}: {', '.join(result['mx_records'])}")

    def start_scan(self, output_file=None):
        self.scan_subdomains()
        
        if output_file:
            self.save_results(output_file)
        
        self.print_summary()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Subdomain Scanner")
    parser.add_argument("domain", help="Target domain name")
    parser.add_argument("--wordlist", help="Path to subdomain wordlist")
    parser.add_argument("--resolvers", help="Comma-separated list of DNS resolvers")
    parser.add_argument("--threads", type=int, default=50,
                      help="Number of threads")
    parser.add_argument("--output", help="Output file for results")
    args = parser.parse_args()

    resolvers = args.resolvers.split(',') if args.resolvers else None
    scanner = SubdomainScanner(args.domain, args.wordlist, resolvers, args.threads)
    scanner.start_scan(args.output)
